Chapter 8:  The Demonstration Programs

(a)  Introduction

The source code and the .EXE files for nine demonstration programs are
provided in two subdirectories:  The \DEMO_EXE subdirectory contains the
executable programs and the \DEMO_SRC subdirectory contains the source
code for these programs.  (FRAG.C generates both FRAG_D.EXE and
FRAG_M.EXE.)  The header file L_MEM.H is necessary for compilation of
the demo programs.  NOMEMLOG.C is also necessary for the compilation of
FRAG.C when DOLPHIN is #defined as TRUE.  (The use of NOMEMLOG.C is
explained in Chapter 9.)  These two files are in the \INCLUDE subdirectory.

The .EXE files may be run directly; some take command line parameters,
as explained in the following sections.  Some of the programs output one
or more memory allocation log reports; these are reproduced in the
corresponding sections below.

The demo programs can be compiled with either Version 5.10 or Version
6.00 of the Microsoft C Compiler.  To compile the demo programs the
environment should be set up for normal compilation with (that is, the
path should allow access the the compiler and linker files, etc.). In
addition:

(1)   L_MEM.H should either be copied into the subdirectory containing
      the Microsoft header files or retained in the same subdirectory
      as the source files.  NOMEMLOG.C should accompany L_MEM.H.

(2)   The Dolphin library files for whichever model(s) you wish to work
      with (small:SMEMM5E.LIB, medium:MMEMM5E.LIB or large:LMEMM5E.LIB)
      should be copied into the subdirectory containing the Microsoft
      library files.  (They can be placed elsewhere if you tell the
      linker where to find them.)

The linker will not search the Dolphin library file during linking,
unless it is told to do so.  This is done by invoking CL with the /link
switch, as in CL DEMO.C /link SMEMM5E.LIB.

If you wish to put L_MEM.H into a special subdirectory, e.g.
F:\C\INCLUDE\DOLPHLIB then the /I compiler switch must be used to
tell the Compiler where to find the header file, as in
/IF:\C\INCLUDE\DOLPHLIB.

The SET command can be used to automate the switches.  E.g. you
might use SET CL=/W3 /IF:\C\INCLUDE\DOLPHLIB /link SMEMM5E.LIB
with the result that all commands of the form CL DEMO.C are executed as
if you had entered CL /W3 /IF:\C\INCLUDE\DOLPHLIB /link SMEMM5E.LIB DEMO.C.


(b)     Explanation of the TINYBUG.C program

The TINYBUG.C program shows how the memory allocation log can be used to
reveal a bug.  The program is as follows:

    /*  TINYBUG.C
     *  this program contains a tiny bug
     *  last mod.: 02-JAN-91
     */

    #include <STDIO.H>

    #include <L_MEM.H>

    Far_str_ptr ptr[1000];

    void main(int argc, char **argv);

    void main(int argc,char **argv)
    {
    int i, err_flag, num_blocks;
    Ulong block_size=2000L;
    Boolean log=FALSE;

    printf("\nUse TINYBUG to run without the memory log"
           "\n or TINYBUG 1 to run with the memory log.\n"
           "\nThis program allocates %lu-byte blocks and frees them.",
            block_size);

    if ( argc > 1 )
       {
       log = TRUE;
       printf("\nA summary log report is then displayed, "
                "and the log is erased.\n");
       }

    printf("\nLargest available block: %lu bytes.",mem_max(&err_flag));

    if ( log )
        mem_log_on();

    /*  allocate memory blocks until insufficient memory  */
    printf("\nNow allocating blocks ... ");
    num_blocks = 0;
    i = 0;
    while ( TRUE )
       {
       ptr[i++] = mem_alloc(block_size,&err_flag);
       if ( err_flag )         /*  if error  */
           break;              /*  exit loop  */
       else                    /*  else  */
           num_blocks++;       /*  increment block count  */
       }
    printf("  %d blocks allocated.",num_blocks);

    /*  now free the blocks  */
    printf("\nNow freeing blocks ... ");

    /*  decrement i to compensate for NULL pointer
     *  returned by last mem_alloc()
     */
    i--;

    while ( num_blocks-- > 0 )
        mem_free(ptr[i--],&err_flag);

    printf("done.");

    if ( log )
       {
       printf("\nNow producing report #3:");
       mem_log_report_3(stdout,0,FALSE);
       printf("\nNow erasing log ... ");
       mem_erase_log(&err_flag);
       printf("done.");
       }

    printf("\nLargest available block: %lu bytes.\n",mem_max(&err_flag));
    }

The program allocates 2000-byte blocks of memory until no more can be
allocated, and then frees the blocks.  When TINYBUG is run with no
command line parameter, so that the log is not used, the screen output
is as follows:

    Use TINYBUG to run without the memory log
     or TINYBUG 1 to run with the memory log.

    This program allocates 2000-byte blocks and frees them.
    Largest available block: 486880 bytes.
    Now allocating blocks ...   241 blocks allocated.
    Now freeing blocks ... done.
    Largest available block: 484864 bytes.

The program runs OK, and there would be no reason to suppose a bug was
lurking except that one might expect that the two values for the largest
available memory block, before and after the allocations and frees,
would be the same, whereas they differ by 2016 bytes.

When TINYBUG is run with command line parameter "1", so that the log is
used, the screen output is as follows:

    Use TINYBUG to run without the memory log
     or TINYBUG 1 to run with the memory log.

    This program allocates 2000-byte blocks and frees them.
    A summary log report is then displayed, and the log is erased.

    Largest available block: 486880 bytes.
    Now allocating blocks ...   235 blocks allocated.
    Now freeing blocks ... done.
    Now producing report #3:
       Number of entries in the memory allocation log: 471
     235 memory blocks (or arrays) were allocated.
     234 blocks (or arrays) were freed.
       1 block (or array) was not freed.
       1 memory operation had insufficient memory.
       1 null pointer passed.

    Now erasing log ... done.
    Largest available block: 484864 bytes.

The summary log report tells us immediately that one of the allocated
blocks was not freed.  Thus we know there is a bug in the following
code:

    /*  decrement i to compensate for NULL pointer
     *  returned by last mem_alloc()
     */
    i--;

    while ( num_blocks-- > 0 )
        mem_free(ptr[i--],&err_flag);

Interested readers may care to identify the bug.  (The answer to the
question is given at the end of this chapter.)


(c)     Explanation of the FRAG.C program

The FRAG.C and the FMALLBUG.C programs demonstrate that Microsoft's far
memory allocation function _fmalloc() (to which malloc() maps in the
medium and large models) has problems.  These programs also demonstrate
that mem_alloc(), Dolphin Software's replacement for _fmalloc() does not
have these problems.  The FRAG.C program does not make us of the memory
allocation log, whereas FMALLBUG.C does.  The FRAG.C program is as
follows:

    /*  FRAG.C
     *  demonstrates failure of Microsoft's _fmalloc()
     *  under certain conditions, and that mem_alloc()
     *  does not fail under the same conditions
     *  last mod.: 11-FEB-91
     */

    #include <STDIO.H>
    #include <STDLIB.H>
    #include <MALLOC.H>
    #include <CONIO.H>
    #include <FCNTL.H>
    #include <IO.H>

    #include <L_MEM.H>

    #define DOLPHIN FALSE    /*  switches between _fmalloc() and mem_alloc()  */

    #if DOLPHIN
    /*  memory log not used in this program so:  */
    #include <NOMEMLOG.C>
    /*  and compile with /NOE linker switch  */
    char *program = "FRAG_D";       /*  rename FRAG.EXE to FRAG_D.EXE  */
    #else
    char *program = "FRAG_M";       /*  rename FRAG.EXE to FRAG_M.EXE  */
    #endif

    #define N       40       /*  number of pointers  */

    void far * ptr[N];

    void main (int argc,char **argv)
    {
    Uint i, j, k, size, n;
    char a[4];
    #if DOLPHIN
    int err_flag;
    #endif

    printf("\nUse \"%s [random_seed [mem_alloc_strategy]]\"."
        "\n    mem_alloc_strategy = 0, 1 or 2"
        "\nThis program allocates blocks of even random size (2 - 65,534 bytes)."
        "\nWhen an attempt to allocate a block of a given size fails,"
        "\nrandomly chosen blocks are freed until that block can be allocated."
        "\nA \"+\" is displayed when an allocation occurs."
        "\nA \"!\" is displayed when an attempt to allocate fails"
        "\n        followed by the number of currently allocated blocks."
        "\nA \"-\" is displayed when a block is freed."
        "\nThe program continues until an allocation fails and"
        "\nthere are no allocated blocks which can be freed."
        #if DOLPHIN
        "\nPress a key to begin (and control-break to quit) ...\n"
        #else
        "\nPress a key to begin ...\n"
        #endif
        ,program);
    getch();

    srand(atoi(argv[1]));
    n = ( argc < 3 ? 0 : argv[2][0]-'0' );
    set_mem_alloc_strategy( n < 3 ? n : 0 );

    #if DOLPHIN
    /*  trap any critical error that occurs  */
    crit_err_termination_on();
    #endif

    /*  initialize pointers to NULL  */
    for ( i=0; i<N; i++ )
        ptr[i] = (void far *)NULL;

    while ( TRUE )
       {
       i = rand()%N;       /*  pick an unused pointer at random  */
       if ( ptr[i] == (void far *)NULL )
          {
          /*  pick an even random size in the range 2 through 65,534 bytes  */
          size = 0;
          while ( !size )
              size = 2*(Uint)rand();
          /*  attempt to allocate a block of this size  */
          while ( TRUE )
             {
             #if DOLPHIN     /*  use Dolphin Software's mem_alloc()  */
             if ( ( ptr[i] = mem_alloc(size,&err_flag) ) != (void far *)NULL )
             #else           /*  use Microsoft's _fmalloc()  */
             if ( ( ptr[i] = _fmalloc(size) ) != (void far *)NULL )
             #endif
                {
                printf("+");    /*  block allocated  */
                break;          /*  go get another unused pointer  */
                }
             else
                {
                printf("!");    /*  allocation attempt failed  */
                /*  see if any blocks available to free  */
                n = 0;
                for ( k=0; k<N; k++ )
                    n += ( ptr[k] != (void far *)NULL );
                /*  display number of blocks currently allocated  */
                printf(itoa(n,a,10));
                if ( n )       /*  if there are blocks left to free  */
                   {
                   /*  free a randomly selected block  */
                   while ( TRUE )
                      {
                      j = rand()%N;
                      if ( ptr[j] != (void far *)NULL )
                         {
                         #if DOLPHIN
                         mem_free(ptr[j],&err_flag);
                         #else
                         _ffree(ptr[j]);
                         #endif
                         printf("-");
                         /*  show pointer as not in use  */
                         ptr[j] = (void far *)NULL;
                         break;  /*  go try the allocation again  */
                         }
                      }
                   }
                else        /*  there are no blocks left to free  */
                   {
                   printf("\nA block of %u bytes could not be allocated, "
                       "even though"
                       "\nall previously allocated blocks had been freed.",
                       size);
                   /*  set memory allocation strategy to normal value  */
                   set_mem_alloc_strategy(FIRST_FIT);
                   exit(1);
                   }
                }
             }
          }
       }
    }

The executable program FRAG_M.EXE (in subdirectory \DEMO_EXE) is the
result of compiling this program with DOLPHIN #defined as FALSE, so that
far memory allocation is done using Microsoft's _fmalloc(). The
executable program FRAG_D.EXE (in the same subdirectory) is the result
of compiling this program with DOLPHIN #defined as TRUE, so that far
memory allocation is done using Dolphin Software's mem_alloc().  (In
this case if NOMEMLOG.C is #included the linker switch /NOE is required,
as explained in Chapter 9.  The #inclusion of NOMEMLOG.C is not stricly
necessary, but shortens the resulting .EXE file.)

Regardless of whether DOLPHIN is #defined as TRUE or as FALSE, the
program functions as follows:  An array of 40 far pointers are
initialized to NULL.  One of these unused pointers is chosen at random.
A size is then chosen, being an even number randomly selected from the
range 2 through 65,564, and an attempt is made to allocate a memory
block of that size.  If successful, the selected pointer is assigned the
address of the allocated block.  This continues until an attempt to
allocate a block fails, due to insufficient memory in the far heap.
When this happens, an allocated block is selected at random and is
freed, and another attempt is made to allocate a block of the selected
size.  Normally after one or a few blocks are freed, there will be
sufficient free memory to allocate the block.

Two parameters may be passed on the command line, a random number seed
and a memory allocation strategy (see Chapter 7, Section (a), for more
on this).  The random number seed modifies the sequence of block sizes
and has a significant effect on the course of the program.  The memory
allocation strategy may have some effect, but not much.  If no values
are passed on the command line then a random number seed of zero, and a
first fit allocation strategy, are used.

In theory this program should continue forever, because on a PC with at
least 256K, by freeing up all blocks there should always be enough
memory in the far heap for a block less than 64K in size to be
allocated.  And in fact FRAG_D, which uses mem_alloc(), does go on
indefinitely.  FRAG_M, however, which uses _fmalloc(), always stops after
a while.  It may go on for quite a while or it may stop immediately, as
in the following case:

    Use "FRAG_M [random_seed [mem_alloc_strategy]]".
        mem_alloc_strategy = 0, 1 or 2
    This program allocates blocks of even random size (2 - 65,534 bytes).
    When an attempt to allocate a block of a given size fails,
    randomly chosen blocks are freed until that block can be allocated.
    "+" is displayed when an allocation occurs.
    "!" is displayed when an attempt to allocate fails
        followed by the number of currently allocated blocks.
    "-" is displayed when a block is freed.
    The program continues until an allocation fails and
    there are no allocated blocks which can be freed.
    Press a key to begin ...
    +++++++++++!11-!10-+++++!14-!13-+!13-!12-!11-!10-!9-!8-!7-!6-!5-!4
    -!3-++++++!8-!7-!6-!5-!4-!3-!2-!1-!0
    A block of 65162 bytes could not be allocated, even though
    all previously allocated blocks had been freed.

We can interpret the output to tell us that 11 blocks were allocated
consecutively, then an attempt at allocation failed.  A block was freed,
but still the allocation failed.  Another block was freed, then the
allocation succeeded, followed by four more allocations, and then
another fail, and so on.  Eventually (and quite quickly in this case) an
attempt at allocation failed, and freeing all allocated blocks was still
not sufficient to allow a successful allocation.  At this point _fmalloc()
could not allocate a block of 65,162 bytes even though there were no
allocated blocks in the far heap.

If some random number is passed to FRAG_M.EXE then this will change the
size of the allocated blocks and thus affect how long the program will
run before it stops.  As FRAG_D.EXE shows, replacing _fmalloc() with
mem_alloc() eliminates the problem, since FRAG_D.EXE does not come to a
halt no matter how long you let it run.


(d)     Explanation of the FMALLBUG.C program

The FMALLBUG.C program is given below:

    /*  FMALLBUG.C
     *  this program reveals a problem with Microsoft's far memory
     *  allocation function, _fmalloc(); the memory allocation
     *  functions in this library do not themselves have this problem;
     *  however, the problem reappears when the memory log is used;
     *  this problem can be avoided by compacting the memory log
     *  before re-allocating memory blocks, as this program shows
     *  last mod.: 02-JAN-91
     */

    #include <STDIO.H>
    #include <STDLIB.H>
    #include <FCNTL.H>
    #include <CONIO.H>
    #include <IO.H>
    #include <MALLOC.H>
    #include <STRING.H>

    #include "L_MEM.H"

    typedef struct mb
       {
       struct mb far *mb_previous;
       int mb_data;
       } Mem_block;

    Mem_block far *list;
    Mem_block far *p;
    Uint first_alloc_size =  40000;
    Uint second_alloc_size = 46000;
    Ulong first_allocation, second_allocation;
    int err_flag, ch;
    char msg[81];

    void main(int argc, char **argv);
    void display_first_allocation(void);
    void display_second_allocation(void);
    Ulong msc_allocation(Uint block_size);
    Ulong dolphin_allocation(Uint block_size, int *err_flag);
    void compact_mem_log(void);
    void print_reports(void);

    void main(int argc,char **argv)
    {
    set_mem_alloc_strategy(atoi(argv[1]));

    printf("\n\nmem_alloc() without memory log (max. avail.: %lu bytes):",
        mem_max(&err_flag));
    first_allocation = dolphin_allocation(first_alloc_size,&err_flag);
    display_first_allocation();
    second_allocation = dolphin_allocation(second_alloc_size,&err_flag);
    display_second_allocation();

    mem_log_on();
    /*  write program name to log  */
    strcpy(msg,"Program is ");
    strcat(msg,argv[0]);
    mem_log_msg(msg,&err_flag);

    printf("\n\nmem_alloc() with memory log (max. avail.: %lu bytes):",
        mem_max(&err_flag));
    first_allocation = dolphin_allocation(first_alloc_size,&err_flag);
    display_first_allocation();
    second_allocation = dolphin_allocation(second_alloc_size,&err_flag);
    display_second_allocation();
    printf("\n");

    compact_mem_log();
    sprintf(msg,"mem_alloc() with memory log and compaction "
        "(max. avail.: %lu bytes)",mem_max(&err_flag));
    printf("\n%s:",msg);
    /*  put the message in the memory log  */
    mem_log_msg(msg,&err_flag);
    first_allocation = dolphin_allocation(first_alloc_size,&err_flag);
    display_first_allocation();
    compact_mem_log();
    second_allocation = dolphin_allocation(second_alloc_size,&err_flag);
    display_second_allocation();
    printf("\n");

    /*  can't do this at the beginning because _fmalloc() ties up memory
     *  (as this program demonstrates) and so mem_alloc() does not succeed
     */
    compact_mem_log();
    printf("\nUsing _fmalloc() (max. available: %lu bytes):",
        mem_max(&err_flag));
    first_allocation = msc_allocation(first_alloc_size);
    display_first_allocation();
    second_allocation = msc_allocation(second_alloc_size);
    display_second_allocation();
    print_reports();
    }

    void display_first_allocation(void)
    {
    Uint num_blocks = (Uint)(first_allocation/first_alloc_size);

    printf("\nUsing block size %u, %lu bytes allocated (%u block%s)%s.",
        first_alloc_size,first_allocation,num_blocks,
        ( num_blocks==1 ? "" : "s" ),
        ( num_blocks==0 ? "" : " and freed" ) );
    }

    void display_second_allocation(void)
    {
    Uint num_blocks = (Uint)(second_allocation/second_alloc_size);

    printf("\nUsing block size %u, %lu bytes allocated (%u block%s)%s.",
        second_alloc_size,second_allocation,num_blocks,
        ( num_blocks==1 ? "" : "s" ),
        ( num_blocks==0 ? "" : " and freed" ) );
    }

    Ulong msc_allocation(Uint block_size)
    {
    Ulong total_allocation = 0L;

    list = (Mem_block far *)NULL;

    /*    allocate memory blocks until memory exhausted  */
    while ( ( p = (Mem_block far *)_fmalloc(block_size) )
        != (Mem_block far *)NULL )
       {
       p->mb_previous = list;
       list = p;
       total_allocation += block_size;
       }

    /*    free the allocated blocks  */
    while ( list != (Mem_block far *)NULL )
       {
       p = list;
       list = list->mb_previous;
       _ffree(p);
       }

    return ( total_allocation );
    }

    Ulong dolphin_allocation(Uint block_size,
                             int *err_flag)
    {
    Ulong total_allocation = 0L;

    list = (Mem_block far *)NULL;
    *err_flag = NO_ERROR;

    /*    allocate memory blocks until memory exhausted  */
    while ( TRUE )
       {
       p = (Mem_block far *)mem_alloc((Ulong)block_size,err_flag);
       if ( *err_flag )
           break;
       p->mb_previous = list;
       list = p;
       total_allocation += block_size;
       }

    if ( *err_flag == INSUFFICIENT_MEMORY )
        *err_flag = NO_ERROR;

    if ( *err_flag )
       {
       printf("\n%s",mem_err_msg(*err_flag,msg));
       exit(1);
       }
    else
       {
       /*    free the allocated blocks  */
       while ( TRUE )
          {
          p = list;
          list = list->mb_previous;
          mem_free((Far_str_ptr)p,err_flag);
          if ( *err_flag == NULL_POINTER )
              break;
          else if ( *err_flag != NO_ERROR )
             {
             printf("\n%s",mem_err_msg(*err_flag,msg));
             exit(2);
             }
          }
       }

    return ( total_allocation );
    }

    void compact_mem_log(void)
    {
    compact_log(&err_flag);
    if ( err_flag )
       {
       printf("\nError when compacting log: %s",mem_err_msg(err_flag,msg));
       mem_log_msg("Error when compacting log.",&err_flag);
       }
    else
       {
       printf("\nMemory log compacted.");
       mem_log_msg("Memory log compacted.",&err_flag);
       }
    }

    void print_reports(void)
    {
    FILE *output;
    char *logfile = "FMALLBUG.LOG";

    printf("\n\nMemory allocation log has %u entries.",mem_log_entries());
    printf("\nPress P to send reports #1 and #2 to printer,"
           "\nF to send reports to disk file %s,"
           "\nany other key to quit ... ",logfile);

    while ( kbhit() )
        getch();
    ch = getche();
    if ( ch == 'P' || ch == 'p' )
       {
       mem_log_report_1(stdprn,1,TRUE,&err_flag);
       mem_log_report_2(stdprn,1,TRUE,&err_flag);
       }
    else if ( ch == 'F' || ch == 'f' )
       {
       if ( ( output = fopen(logfile,"wt") ) == NULL )
          {
          printf("\nUnable to open file %s",logfile);
          exit(1);
          }
       mem_log_report_1(output,1,TRUE,&err_flag);
       if ( err_flag )
           printf("\nError when writing first report: %s.",
               mem_err_msg(err_flag,msg));
       else
          {
          mem_log_report_2(output,1,TRUE,&err_flag);
          if ( err_flag )
              printf("\nError when writing second report: %s.",
              mem_err_msg(err_flag,msg));
          }
       }

    printf("\n");
    }

This program demonstrates an unreliable aspect of Microsoft's _fmalloc()
function.  If a bunch of memory blocks, in this case 40,000-byte blocks,
are allocated with _fmalloc() and then freed, there should be enough
free memory to allocate another bunch of 46,000-byte blocks.  In fact
_fmalloc() will be able to allocate at most two such blocks, despite the
fact that most of the far heap is not in use.

FMALLBUG.C demonstrates this, in the latter part of the program.  The
earlier part of the program shows that mem_alloc() - since it is based
on the DOS memory allocation service, does not have this problem.  A
typical screen output will begin thus:

    mem_alloc() without memory log (max. avail.: 483520 bytes):
    Using block size 40000, 480000 bytes allocated (12 blocks) and freed.
    Using block size 46000, 460000 bytes allocated (10 blocks) and freed.

However, if the memory log is turned on then the problem reappears:

    mem_alloc() with memory log (max. avail.: 483280 bytes):
    Using block size 40000, 480000 bytes allocated (12 blocks) and freed.
    Using block size 46000, 0 bytes allocated (0 blocks).

This is because each log entry takes up 32 bytes of allocated memory in
the far heap, thus causing the far heap to become fragmented.  The
problem may be overcome by using the Dolphin library compact_log()
function (in the compact_mem_log() function in FMALLBUG.C).  This takes
the log entries, scattered about the far heap, and consolidates them
into one contiguous part of memory (or occasionally into several areas
of memory).  With compaction we obtain:

    Memory log compacted.
    mem_alloc() with memory log and compaction (max. avail.: 481648 bytes):
    Using block size 40000, 480000 byFtes allocated (12 blocks) and freed.
    Memory log compacted.
    Using block size 46000, 460000 bytes allocated (10 blocks) and freed.

Finally _fmalloc() is used (it could not be used before mem_alloc()
because the far heap fragmentation would prevent mem_alloc() from
allocating memory properly):

    Memory log compacted.
    Using _fmalloc() (max. available: 475312 bytes):
    Using block size 40000, 440000 bytes allocated (11 blocks) and freed.
    Using block size 46000, 46000 bytes allocated (1 block) and freed.

Even though all eleven blocks allocated using _fmalloc() were freed,
thus freeing a total of 440,000 bytes, _fmalloc() can only find space to
allocate a single 46,000-byte block.

The FMALLBUG program gives you the option of sending the memory
allocation log to printer or to disk file.  The log should look
something like the following:

    Number of entries in the memory allocation log: 84

       1  Message: Program is E:\MEMLIB\DEMOS\FMALLBUG.EXE

       2 "Ascertain maximum available."  483280 bytes.

       3 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 29FD:0000.

       4 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 33C5:0000.

       5 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 3D8D:0000.

       6 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 4755:0000.

       7 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 511D:0000.

       8 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 5AE5:0000.

       9 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 64AD:0000.

      10 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 6E75:0000.

      11 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 783D:0000.

      12 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 8205:0000.

      13 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 8BCD:0000.

      14 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 9595:0000.

      15 "Allocate 40000 bytes."
         Error: Insufficient memory.  Maximum available is 2464 bytes.

      16 "Free block at 9595:0000."  40000 bytes (0x9C4 paragraphs) freed.

      17 "Free block at 8BCD:0000."  40000 bytes (0x9C4 paragraphs) freed.

      18 "Free block at 8205:0000."  40000 bytes (0x9C4 paragraphs) freed.

      19 "Free block at 783D:0000."  40000 bytes (0x9C4 paragraphs) freed.

      20 "Free block at 6E75:0000."  40000 bytes (0x9C4 paragraphs) freed.

      21 "Free block at 64AD:0000."  40000 bytes (0x9C4 paragraphs) freed.

      22 "Free block at 5AE5:0000."  40000 bytes (0x9C4 paragraphs) freed.

      23 "Free block at 511D:0000."  40000 bytes (0x9C4 paragraphs) freed.

      24 "Free block at 4755:0000."  40000 bytes (0x9C4 paragraphs) freed.

      25 "Free block at 3D8D:0000."  40000 bytes (0x9C4 paragraphs) freed.

      26 "Free block at 33C5:0000."  40000 bytes (0x9C4 paragraphs) freed.

      27 "Free block at 29FD:0000."  40000 bytes (0x9C4 paragraphs) freed.

      28 "Free block at 0:0000."
         Error: Null pointer.

      29 "Allocate 46000 bytes."
         Error: Insufficient memory.  Maximum available is 39808 bytes.

      30 "Free block at 0:0000."
         Error: Null pointer.

      31  Message: Memory log compacted.

      32 "Ascertain maximum available."  481648 bytes.

      33  Message: mem_alloc() with memory log and compaction (max. avail.: 481648 bytes)

      34 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 2A63:0000.

      35 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 342B:0000.

      36 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 3DF3:0000.

      37 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 47BB:0000.

      38 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 5183:0000.

      39 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 5B4B:0000.

      40 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 6513:0000.

      41 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 6EDB:0000.

      42 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 78A3:0000.

      43 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 826B:0000.

      44 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 8C33:0000.

      45 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 95FB:0000.

      46 "Allocate 40000 bytes."
         Error: Insufficient memory.  Maximum available is 832 bytes.

      47 "Free block at 95FB:0000."  40000 bytes (0x9C4 paragraphs) freed.

      48 "Free block at 8C33:0000."  40000 bytes (0x9C4 paragraphs) freed.

      49 "Free block at 826B:0000."  40000 bytes (0x9C4 paragraphs) freed.

      50 "Free block at 78A3:0000."  40000 bytes (0x9C4 paragraphs) freed.

      51 "Free block at 6EDB:0000."  40000 bytes (0x9C4 paragraphs) freed.

      52 "Free block at 6513:0000."  40000 bytes (0x9C4 paragraphs) freed.

      53 "Free block at 5B4B:0000."  40000 bytes (0x9C4 paragraphs) freed.

      54 "Free block at 5183:0000."  40000 bytes (0x9C4 paragraphs) freed.

      55 "Free block at 47BB:0000."  40000 bytes (0x9C4 paragraphs) freed.

      56 "Free block at 3DF3:0000."  40000 bytes (0x9C4 paragraphs) freed.

      57 "Free block at 342B:0000."  40000 bytes (0x9C4 paragraphs) freed.

      58 "Free block at 2A63:0000."  40000 bytes (0x9C4 paragraphs) freed.

      59 "Free block at 0:0000."
         Error: Null pointer.

      60  Message: Memory log compacted.

      61 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 2B17:0000.

      62 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 3653:0000.

      63 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 418F:0000.

      64 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 4CCB:0000.

      65 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 5807:0000.

      66 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 6343:0000.

      67 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 6E7F:0000.

      68 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 79BB:0000.

      69 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 84F7:0000.

      70 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 9033:0000.

      71 "Allocate 46000 bytes."
         Error: Insufficient memory.  Maximum available is 18560 bytes.

      72 "Free block at 9033:0000."  46000 bytes (0xB3B paragraphs) freed.

      73 "Free block at 84F7:0000."  46000 bytes (0xB3B paragraphs) freed.

      74 "Free block at 79BB:0000."  46000 bytes (0xB3B paragraphs) freed.

      75 "Free block at 6E7F:0000."  46000 bytes (0xB3B paragraphs) freed.

      76 "Free block at 6343:0000."  46000 bytes (0xB3B paragraphs) freed.

      77 "Free block at 5807:0000."  46000 bytes (0xB3B paragraphs) freed.

      78 "Free block at 4CCB:0000."  46000 bytes (0xB3B paragraphs) freed.

      79 "Free block at 418F:0000."  46000 bytes (0xB3B paragraphs) freed.

      80 "Free block at 3653:0000."  46000 bytes (0xB3B paragraphs) freed.

      81 "Free block at 2B17:0000."  46000 bytes (0xB3B paragraphs) freed.

      82 "Free block at 0:0000."
         Error: Null pointer.

      83  Message: Memory log compacted.

      84 "Ascertain maximum available."  475312 bytes.

       Number of entries in the memory allocation log: 84

       ***  Block/array ID = 1  ***

       3 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 29FD:0000.

      27 "Free block at 29FD:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 2  ***

       4 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 33C5:0000.

      26 "Free block at 33C5:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 3  ***

       5 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 3D8D:0000.

      25 "Free block at 3D8D:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 4  ***

       6 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 4755:0000.

      24 "Free block at 4755:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 5  ***

       7 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 511D:0000.

      23 "Free block at 511D:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 6  ***

       8 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 5AE5:0000.

      22 "Free block at 5AE5:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 7  ***

       9 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 64AD:0000.

      21 "Free block at 64AD:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 8  ***

      10 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 6E75:0000.

      20 "Free block at 6E75:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 9  ***

      11 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 783D:0000.

      19 "Free block at 783D:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 10  ***

      12 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 8205:0000.

      18 "Free block at 8205:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 11  ***

      13 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 8BCD:0000.

      17 "Free block at 8BCD:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 12  ***

      14 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 9595:0000.

      16 "Free block at 9595:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 13  ***

      34 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 2A63:0000.

      58 "Free block at 2A63:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 14  ***

      35 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 342B:0000.

      57 "Free block at 342B:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 15  ***

      36 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 3DF3:0000.

      56 "Free block at 3DF3:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 16  ***

      37 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 47BB:0000.

      55 "Free block at 47BB:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 17  ***

      38 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 5183:0000.

      54 "Free block at 5183:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 18  ***

      39 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 5B4B:0000.

      53 "Free block at 5B4B:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 19  ***

      40 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 6513:0000.

      52 "Free block at 6513:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 20  ***

      41 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 6EDB:0000.

      51 "Free block at 6EDB:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 21  ***

      42 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 78A3:0000.

      50 "Free block at 78A3:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 22  ***

      43 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 826B:0000.

      49 "Free block at 826B:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 23  ***

      44 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 8C33:0000.

      48 "Free block at 8C33:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 24  ***

      45 "Allocate 40000 bytes."
         2500 (0x9C4) paragraphs (40000 bytes) allocated at 95FB:0000.

      47 "Free block at 95FB:0000."  40000 bytes (0x9C4 paragraphs) freed.


       ***  Block/array ID = 25  ***

      61 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 2B17:0000.

      81 "Free block at 2B17:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 26  ***

      62 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 3653:0000.

      80 "Free block at 3653:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 27  ***

      63 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 418F:0000.

      79 "Free block at 418F:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 28  ***

      64 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 4CCB:0000.

      78 "Free block at 4CCB:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 29  ***

      65 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 5807:0000.

      77 "Free block at 5807:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 30  ***

      66 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 6343:0000.

      76 "Free block at 6343:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 31  ***

      67 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 6E7F:0000.

      75 "Free block at 6E7F:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 32  ***

      68 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 79BB:0000.

      74 "Free block at 79BB:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 33  ***

      69 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 84F7:0000.

      73 "Free block at 84F7:0000."  46000 bytes (0xB3B paragraphs) freed.


       ***  Block/array ID = 34  ***

      70 "Allocate 46000 bytes."
         2875 (0xB3B) paragraphs (46000 bytes) allocated at 9033:0000.

      72 "Free block at 9033:0000."  46000 bytes (0xB3B paragraphs) freed.


       Errors not specific to a memory block or array:

      15 "Allocate 40000 bytes."
         Error: Insufficient memory.  Maximum available is 2464 bytes.

      28 "Free block at 0:0000."
         Error: Null pointer.

      29 "Allocate 46000 bytes."
         Error: Insufficient memory.  Maximum available is 39808 bytes.

      30 "Free block at 0:0000."
         Error: Null pointer.

      46 "Allocate 40000 bytes."
         Error: Insufficient memory.  Maximum available is 832 bytes.

      59 "Free block at 0:0000."
         Error: Null pointer.

      71 "Allocate 46000 bytes."
         Error: Insufficient memory.  Maximum available is 18560 bytes.

      82 "Free block at 0:0000."
         Error: Null pointer.

      34 memory blocks (or arrays) were allocated.
      34 blocks (or arrays) were freed.
       4 memory operations had insufficient memory.
       4 null pointers passed.


(e)     Explanation of the CRIT_ERR.C program

The CRIT_ERR.C program is given below:

    /*  CRIT_ERR.C
     *  demonstrates special termination when a critical error occurs
     *  last mod.: 26-DEC-90
     */

    #include <STDIO.H>
    #include <STDLIB.H>
    #include <STRING.H>
    #include <CONIO.H>

    #include "L_MEM.H"

    char msg[81];

    void main(int argc, char **argv);
    Far_str_ptr allocate(Ulong n);

    void main(int argc,char **argv)
    {
    int err_flag;
    Far_str_ptr ptr1, ptr2, ptr3;

    if ( argc == 1 || !( argv[1][0] == '0' || argv[1][0] == '1' ) )
       {
       printf("\nUse \"CRIT_ERR 0\" to run this program with the "
           "\nmemory allocation log off.\n"
           "\nUse \"CRIT_ERR 1\" to run this program with the log on;"
           "\nthe log will be dumped to the screen if a critical "
           "\nerror occurs.\n"
           "\nYou will be asked permission to overwrite a control block"
           "\nto demonstrate automatic output of the log when such an"
           "\nerror is detected.\n");
       exit(0);
       }

    set_mem_alloc_strategy(FIRST_FIT);

    /*  set critical error termination on
     *  regardless of whether the log is on
     */
    crit_err_termination_on();

    if ( argv[1][0] == '1' )
       {
       mem_log_on();
       /*  specify that report #1 to be output to screen
        *  if a critical error is detected
        */
       set_crit_err_termination_output(SCREEN_OUTPUT,REPORT_1_ONLY);
       }

    /*  allocate a few blocks of 10000 bytes  */

    ptr1 = allocate(7000L);
    ptr2 = allocate(5000L);
    ptr3 = allocate(3000L);

    printf("\n\nWARNING:  For the purpose of demonstrating program termination"
        "\non detection of control block damage, a control block will now"
        "\nbe overwritten deliberately.  This will probably cause the computer"
        "\nto lock up and a DOS error message to appear.  This is no cause for "
        "\nalarm but you will probably have to reboot the system."
        "\nPress Enter to overwrite the control block, any other to skip.\n");

    while ( kbhit() )
        getch();
    if ( getch() == '\xD' )
       {
       memset_f(ptr2,0x55,6000L);
       /*  write 6000 bytes, overwriting a control block  */
       printf("\nAttempting to free the third block produces this message:");
       }

    /*  attempt to free the third block  */
    mem_free(ptr3,&err_flag);

    if ( err_flag )
        printf("\n%s",mem_err_msg(err_flag,msg));
        /*  should get a control-blocks_destroyed error message  */
    else
        printf("\nBlock at %lp freed.\n",ptr3);
        /*  should appear only if control block not overwritten  */
    }

    Far_str_ptr allocate(Ulong n)
    {
    int err_flag;
    Far_str_ptr ptr = mem_alloc(n,&err_flag);

    if ( err_flag )
        printf("\n%s",mem_err_msg(err_flag,msg));
    else
        printf("\n%lu bytes allocated at %lp",n,ptr);

    return ( ptr );
    }

This program deliberately creates a critical memory error and
demonstrates how to trap it and to cause the memory log reports to be
generated.  The critical error is created by first allocating several
memory blocks and then using memset_f() to overwrite one of the memory
control blocks.  Normally a critical error of this sort might or might
not cause the system to lock up.  Sometimes a program may just run on,
doing unintended things.  The function crit_err_termination_on() is used
to set a flag so that the program will terminate as soon as one of the
library functions detects a critical error.  Setting this flag is
independent of whether the memory log is in use.

When the program is run with no command line parameters the
following appears:

    Use "CRIT_ERR 0" to run this program with the memory
    allocation log off.

    Use "CRIT_ERR 1" to run this program with the log on;
    the log will be dumped to the screen if a critical
    error occurs.

    You will be asked permission to overwrite a control block
    to demonstrate automatic output of the log when such an
    error is detected.

If you enter "CRIT_ERR 0" then the log is not turned on, and (if you
give permission) the system will lock up when the statement
mem_free(ptr3,&err_flag) is executed.  If you enter "CRIT_ERR 1" then
the log is turned on, and when the critical error is detected the log is
displayed on the screen.  The screen output should be something like:

    7000 bytes allocated at 292B:0000
    5000 bytes allocated at 2AE5:0000
    3000 bytes allocated at 2C22:0000

    WARNING:  For the purpose of demonstrating program termination
    on detection of control block damage, a control block will now
    be overwritten deliberately.  This will probably cause the computer
    to lock up and a DOS error message to appear.  This is no cause for
    alarm but you will probably have to reboot the system.
    Press Enter to overwrite the control block, any other to skip.

    Attempting to free the third block produces this message:
    Memory control block(s) destroyed.  Log contains 3 entries.

    Number of entries in the memory allocation log: 3

       1 "Allocate 7000 bytes."
         438 (0x1B6) paragraphs (7008 bytes) allocated at 292B:0000.
       2 "Allocate 5000 bytes."
         313 (0x139) paragraphs (5008 bytes) allocated at 2AE5:0000.

    Log entry overwritten.

    DOS error message:
    Memory allocation error
    Cannot load COMMAND, system halted

At this point the computer must be rebooted.  If the statement:

    set_crit_err_termination_output(SCREEN_OUTPUT,REPORT_1_ONLY);

is replaced by:

    set_crit_err_termination_output(PRINTER_OUTPUT,REPORTS_1_AND_2);

then when the critical error is detected report #1 and report #2 will
both be output to the printer, instead of simply report #1 to the
screen.  In this way these reports, giving information concerning memory
operations just prior to the detection of a critical error, can help in
tracking down memory bugs.


(f)     Explanation of the FREE_ALL.C program

The FREE_ALL.C program is as follows:

    /*  FREE_ALL.C
     *  this demonstrates how to free a clutch
     *  of allocated memory blocks in one operation
     *  last mod.: 26-DEC-90
     */

    #include <STDIO.H>
    #include <STDLIB.H>
    #include <STRING.H>
    #include <FCNTL.H>
    #include <CONIO.H>
    #include <IO.H>

    #include "L_MEM.H"

    #define MAX_SIZE 153600L

    int err_flag;
    char msg[50];

    void main(int argc, char **argv);
    Uint exhaust_memory(void);
    void print_error(void);
    void print_reports(void);

    void main(int argc,char **argv)
    {
    Far_str_ptr ptr1, ptr2, ptr3;

    srand(atoi(argv[1]));

  printf("\nUse:  FREE_ALL <random_seed>"
  "\nThis program demonstrates the use of mem_free_all() to free a clutch of"
  "\nmemory blocks.  First some memory blocks are allocated.  Then memory is "
  "\nallocated in blocks of random size (up to %dK) until an \"insufficient"
  "\nmemory\" error is encountered.  The second set of blocks are then all"
  "\nfreed at once using mem_free_all().  The previous blocks are resized and"
  "\nare then freed, again using mem_free_all().\n",
  MAX_SIZE/1024);

    printf("\nLargest block available = %lu",mem_max(&err_flag));

    /*  turn the memory allocation log on  */
    mem_log_on();
    /*  terminate if control blocks destroyed  */
    crit_err_termination_on();
    /*  on critical error send both reports to printer  */
    set_crit_err_termination_output(PRINTER_OUTPUT,REPORT_1_ONLY);

    printf("\n3 memory blocks allocated.");
    ptr1 = mem_alloc(12000L,&err_flag);
    if ( !err_flag )       /*  if no error  */
       {
       ptr2 = mem_alloc(15000L,&err_flag);
       if ( !err_flag )
           ptr3 = mem_alloc_ch(5000L,0xFF,&err_flag);
           /*  omit "ptr3 = " and change this to
            *  mem_alloc_ch(5000L,0xFF,&err_flag);
            *  to create a situation in which a critical error will
            *  (probably occur) causing reports to be sent to the printer
            */
       }
    if ( err_flag )
       {
       printf("\nProblem allocating memory.");
       print_error();
       exit(1);
       }

    /*  call a function in which memory blocks are allocated
     *  and later freed by a call to mem_free_all()
     */
    printf("\n%d more memory blocks allocated ... ",exhaust_memory());
    printf("and freed.");

    /*  resize previously allocated memory blocks  */
    ptr1 = mem_increase(ptr1,20000,&err_flag);
    ptr2 = mem_decrease(ptr2,10000,&err_flag);
    ptr3 = mem_resize_ch(ptr3,10000L,0xAA,&err_flag);
    printf("\nFirst 3 blocks resized ... ");

    /*  now free up all allocated memory  */
    mem_free_all(0,&err_flag);
    if ( !err_flag )
        printf("and freed.");

    printf("\n\nThis is report #3 (summary from report #2):");
    mem_log_report_3(stdout,0,FALSE);
    print_reports();

    printf("\nLargest block available = %lu\nNow erasing log ...",
        mem_max(&err_flag));

    /*  now throw away the memory log  */
    mem_erase_log(&err_flag);

    if ( err_flag )
       {
       printf("\nWhen erasing memory log: ");
       print_error();
       exit(2);
       }

    printf("\nLargest block available = %lu\n",mem_max(&err_flag));
    }

    /*  perform a bunch of memory allocations until memory is exhausted  */
    Uint exhaust_memory(void)
    {
    Uint n = 0;
    int block_id = get_current_block_id();
    /*  save ID of most recently allocated block
     *  so mem_free_all() knows where to free from
     */

    mem_log_msg("Now allocating blocks until no more space ...",&err_flag);
    err_flag = NO_ERROR;
    while ( err_flag != INSUFFICIENT_MEMORY )
       {
       mem_alloc_ch((((Ulong)rand())*rand())%MAX_SIZE,(char)rand(),&err_flag);
       if ( err_flag == NO_ERROR )
           n++;
       if ( err_flag != INSUFFICIENT_MEMORY && err_flag != NO_ERROR )
          {
          print_error();
          mem_log_report_2(stdout,1,0);
          exit(3);
          }
       }

    /*  now free all blocks allocated using exhaust_memory()
     *  with a call to mem_free_all()
     */
    mem_free_all(block_id,&err_flag);
    if ( err_flag )
       {
       printf("\nWhen freeing allocated blocks: ");
       print_error();
       exit(4);
       }

    return ( n );   /*  number of blocks allocated (and freed)  */
    }

    void print_error(void)
    {
    if ( err_flag )
        printf("\n%s",mem_err_msg(err_flag,msg));
    }

    void print_reports(void)
    {
    int ch;
    FILE *output;
    char *logfile = "FREE_ALL.LOG";

    printf("\nPress P to send reports #1 and #2 to printer,"
           "\nF to send reports to disk file %s,"
           "\nany other key to quit ... ",logfile);

    while ( kbhit() )
        getch();
    ch = getche();
    if ( ch == 'P' || ch == 'p' )
       {
       setmode(fileno(stdprn),O_TEXT);
       mem_log_report_1(stdprn,1,TRUE);
       mem_log_report_2(stdprn,1,TRUE);
       }
    else if ( ch == 'F' || ch == 'f' )
       {
       if ( ( output = fopen(logfile,"wt") ) == NULL )
          {
          printf("\nUnable to open file %s",logfile);
          exit(5);
          }
       mem_log_report_1(output,1,TRUE);
       mem_log_report_2(output,1,TRUE);
       }

    printf("\n");
    }

This program demonstrates the technique whereby, using the memory
allocation log, it is possible to free a clutch of allocated memory
blocks in a single operation.  This technique might be employed, for
example, to free all memory allocated within a particular function, and
this case is demonstrated in this program.  Notice that upon entry to
the function exhaust_memory() the function get_current_block_id() is
called to obtain the ID of the most-recently allocated memory block.
Then, just prior to leaving the function the statement

                mem_free_all(block_id,&err_flag);

is executed.  This has the effect of freeing all blocks that were
allocated since the block with ID block_id, or in other words, of
freeing all blocks that were allocated within the function
exhaust_memory().
   
FREE_ALL produces the following output to the screen:

    Use:  FREE_ALL <random_seed>
    This program demonstrates the use of mem_free_all() to free a clutch of
    memory blocks.  First some memory blocks are allocated.  Then memory is
    allocated in blocks of random size (up to 150K) until an "insufficient
    memory" error is encountered.  The second set of blocks are then all
    freed at once using mem_free_all().  The previous blocks are resized and
    are then freed, again using mem_free_all().

    Largest block available = 476832
    3 memory blocks allocated.
    5 more memory blocks allocated ... and freed.
    First 3 blocks resized ... and freed.

    This is report #3 (summary from report #2):
       Number of entries in the memory allocation log: 25
       8 memory blocks (or arrays) were allocated.
       8 blocks (or arrays) were freed.
       1 memory operation had insufficient memory.

    Press P to send reports #1 and #2 to printer,
    F to send reports to disk file FREE_ALL.LOG,
    any other key to quit ...

    Largest block available = 117168
    Now erasing log ...
    Largest block available = 476832

The memory allocation log which is sent to the printer or to a disk file
should look something like this:

    Number of entries in the memory allocation log: 25

       1 "Allocate 12000 bytes."
         750 (0x2EE) paragraphs (12000 bytes) allocated at 2B96:0000.

       2 "Allocate 15000 bytes."
         938 (0x3AA) paragraphs (15008 bytes) allocated at 2E88:0000.

       3 "Allocate 5000 bytes."  Bytes set to ascii 255 (0xFF).
         313 (0x139) paragraphs (5008 bytes) allocated at 3236:0000.

       4  Message: Now allocating blocks until no more space ...

       5 "Allocate 44922 bytes."  Bytes set to ascii 38 (0x26).
         2808 (0xAF8) paragraphs (44928 bytes) allocated at 337A:0000.

       6 "Allocate 14435 bytes."  Bytes set to ascii 133 (0x85).
         903 (0x387) paragraphs (14448 bytes) allocated at 3E76:0000.

       7 "Allocate 72650 bytes."  Bytes set to ascii 173 (0xAD).
         4541 (0x11BD) paragraphs (72656 bytes) allocated at 4201:0000.

       8 "Allocate 117300 bytes."  Bytes set to ascii 148 (0x94).
         7332 (0x1CA4) paragraphs (117312 bytes) allocated at 53C2:0000.

       9 "Allocate 87697 bytes."  Bytes set to ascii 118 (0x76).
         5482 (0x156A) paragraphs (87712 bytes) allocated at 706A:0000.

      10 "Allocate 111325 bytes."
         Error: Insufficient memory.  Maximum available is 106992 bytes.

      11  Message: About to free all allocated blocks/arrays with ID > 3.

      12 "Free block at 337A:0000."  44928 bytes (0xAF8 paragraphs) freed.

      13 "Free block at 3E76:0000."  14448 bytes (0x387 paragraphs) freed.

      14 "Free block at 4201:0000."  72656 bytes (0x11BD paragraphs) freed.

      15 "Free block at 53C2:0000."  117312 bytes (0x1CA4 paragraphs) freed.

      16 "Free block at 706A:0000."  87712 bytes (0x156A paragraphs) freed.

      17 "Free all blocks/arrays with ID > 3."  5 freed.

      18 "Increase 2B96:0000 by 20000 bytes (from 12000 bytes)."
          2000 (0x7D0) paragraphs (32000 bytes) allocated at 338C:0000.

      19 "Decrease 2E88:0000 by 10000 bytes (from 15008 bytes)."
          313 (0x139) paragraphs (5008 bytes) allocated at same address.

      20 "Resize 3236:0000 to 10000 bytes (from 5008 bytes)."
          Extra bytes set to ascii 170 (0xAA).
          625 (0x271) paragraphs (10000 bytes) allocated at 2B9C:0000.

      21  Message: About to free all allocated blocks/arrays.

      22 "Free block at 338C:0000."  32000 bytes (0x7D0 paragraphs) freed.

      23 "Free block at 2E88:0000."  5008 bytes (0x139 paragraphs) freed.

      24 "Free block at 2B9C:0000."  10000 bytes (0x271 paragraphs) freed.

      25 "Free all blocks/arrays."  3 freed.

       Number of entries in the memory allocation log: 25

       ***  Block/array ID = 1  ***

       1 "Allocate 12000 bytes."
         750 (0x2EE) paragraphs (12000 bytes) allocated at 2B96:0000.

      18 "Increase 2B96:0000 by 20000 bytes (from 12000 bytes)."
          2000 (0x7D0) paragraphs (32000 bytes) allocated at 338C:0000.

      22 "Free block at 338C:0000."  32000 bytes (0x7D0 paragraphs) freed.


       ***  Block/array ID = 2  ***

       2 "Allocate 15000 bytes."
         938 (0x3AA) paragraphs (15008 bytes) allocated at 2E88:0000.

      19 "Decrease 2E88:0000 by 10000 bytes (from 15008 bytes)."
          313 (0x139) paragraphs (5008 bytes) allocated at same address.

      23 "Free block at 2E88:0000."  5008 bytes (0x139 paragraphs) freed.


       ***  Block/array ID = 3  ***

       3 "Allocate 5000 bytes."  Bytes set to ascii 255 (0xFF).
         313 (0x139) paragraphs (5008 bytes) allocated at 3236:0000.

      20 "Resize 3236:0000 to 10000 bytes (from 5008 bytes)."
          Extra bytes set to ascii 170 (0xAA).
          625 (0x271) paragraphs (10000 bytes) allocated at 2B9C:0000.

      24 "Free block at 2B9C:0000."  10000 bytes (0x271 paragraphs) freed.


       ***  Block/array ID = 4  ***

       5 "Allocate 44922 bytes."  Bytes set to ascii 38 (0x26).
         2808 (0xAF8) paragraphs (44928 bytes) allocated at 337A:0000.

      12 "Free block at 337A:0000."  44928 bytes (0xAF8 paragraphs) freed.


       ***  Block/array ID = 5  ***

       6 "Allocate 14435 bytes."  Bytes set to ascii 133 (0x85).
         903 (0x387) paragraphs (14448 bytes) allocated at 3E76:0000.

      13 "Free block at 3E76:0000."  14448 bytes (0x387 paragraphs) freed.


       ***  Block/array ID = 6  ***

       7 "Allocate 72650 bytes."  Bytes set to ascii 173 (0xAD).
         4541 (0x11BD) paragraphs (72656 bytes) allocated at 4201:0000.

      14 "Free block at 4201:0000."  72656 bytes (0x11BD paragraphs) freed.


       ***  Block/array ID = 7  ***

       8 "Allocate 117300 bytes."  Bytes set to ascii 148 (0x94).
         7332 (0x1CA4) paragraphs (117312 bytes) allocated at 53C2:0000.

      15 "Free block at 53C2:0000."  117312 bytes (0x1CA4 paragraphs) freed.


       ***  Block/array ID = 8  ***

       9 "Allocate 87697 bytes."  Bytes set to ascii 118 (0x76).
         5482 (0x156A) paragraphs (87712 bytes) allocated at 706A:0000.

      16 "Free block at 706A:0000."  87712 bytes (0x156A paragraphs) freed.


       Errors not specific to a memory block or array:

      10 "Allocate 111325 bytes."
         Error: Insufficient memory.  Maximum available is 106992 bytes.

       8 memory blocks (or arrays) were allocated.
       8 blocks (or arrays) were freed.
       1 memory operation had insufficient memory.

FREE_ALL provides an example of how the memory log can help in
debugging.  During the development of this demo program the author
originally wrote the statement:

            ptr3 = mem_alloc_ch(5000L,0xFF,&err_flag);

as:         mem_alloc_ch(5000L,0xFF,&err_flag);

and (unlike the statements for the previous two allocations) it was not
modified.  The effect on the program is that ptr3 is not initialized but
is later passed to mem_resize_ch() in the statement:

            ptr3 = mem_resize_ch(ptr3,10000L,0xAA,&err_flag);

This is sufficient to cause the program to crash (although it might
behave differently on different systems).  Since the statements:

    crit_err_termination_on();
    set_crit_err_termination_output(PRINTER_OUTPUT,REPORT_1_ONLY);

are called at the start of the program, when the fatal call to
mem_resize_ch() occurs, and a problem is detected, report #1 is sent to
the printer, as follows:

    Number of entries in the memory allocation log: 20

       1 "Allocate 12000 bytes."
         750 (0x2EE) paragraphs (12000 bytes) allocated at 2B95:0000.

       2 "Allocate 15000 bytes."
         938 (0x3AA) paragraphs (15008 bytes) allocated at 2E87:0000.

       3 "Allocate 5000 bytes."  Bytes set to ascii 255 (0xFF).
         313 (0x139) paragraphs (5008 bytes) allocated at 3235:0000.

       4  Message: Now allocating blocks until no more space ...

       5 "Allocate 44922 bytes."  Bytes set to ascii 38 (0x26).
         2808 (0xAF8) paragraphs (44928 bytes) allocated at 3379:0000.

       6 "Allocate 14435 bytes."  Bytes set to ascii 133 (0x85).
         903 (0x387) paragraphs (14448 bytes) allocated at 3E75:0000.

       7 "Allocate 72650 bytes."  Bytes set to ascii 173 (0xAD).
         4541 (0x11BD) paragraphs (72656 bytes) allocated at 4200:0000.

       8 "Allocate 117300 bytes."  Bytes set to ascii 148 (0x94).
         7332 (0x1CA4) paragraphs (117312 bytes) allocated at 53C1:0000.

       9 "Allocate 87697 bytes."  Bytes set to ascii 118 (0x76).
         5482 (0x156A) paragraphs (87712 bytes) allocated at 7069:0000.

      10 "Allocate 111325 bytes."
         Error: Insufficient memory.  Maximum available is 107008 bytes.

      11  Message: About to free all allocated blocks/arrays with ID > 3.

      12 "Free block at 3379:0000."  44928 bytes (0xAF8 paragraphs) freed.

      13 "Free block at 3E75:0000."  14448 bytes (0x387 paragraphs) freed.

      14 "Free block at 4200:0000."  72656 bytes (0x11BD paragraphs) freed.

      15 "Free block at 53C1:0000."  117312 bytes (0x1CA4 paragraphs) freed.

      16 "Free block at 7069:0000."  87712 bytes (0x156A paragraphs) freed.

      17 "Free all blocks/arrays with ID > 3."  5 freed.

      18 "Increase 2B95:0000 by 20000 bytes (from 12000 bytes)."
          2000 (0x7D0) paragraphs (32000 bytes) allocated at 338B:0000.

      19 "Decrease 2E88:0000 by 10000 bytes (from 15008 bytes)."
          313 (0x139) paragraphs (5008 bytes) allocated at same address.

      20 "Resize B73:0000 to 10000 bytes."
         Error: Memory control block apparently overwritten.

The address in log entry 20, B73:0000, is clearly spurious.  Checking
back through the log we find that it should be 3235:0000, the address of
the third block allocated.  Upon checking the program at the point where
the third block is allocated, the error (the omitted "ptr3 = ") becomes
clear.

When B73:0000 is passed to mem_resize_ch() it looks at the memory block
pointed to and, finding an area of memory in which all the bytes have
the same value, it concludes that the memory control block has been
overwritten, whereupon it pulls the emergency cord and the program
terminates forthwith (after sending the report to the printer).


(g)     Explanation of the MEMSTRAT.C program

The MEMSTRAT.C program is as follows:

    /*  MEMSTRAT.C
     *  gets and sets memory allocation strategy
     *  last mod.: 10-DEC-90
     */

    #include <STDIO.H>

    #include "L_MEM.H"

    char *str = "Memory allocation strategy = ";

    void main(int argc,char **argv)
    {
    if ( argc > 1 )
        set_mem_alloc_strategy(argv[1][0]-'0');
    while ( *str )
        putchar(*(str++));
    putchar(get_mem_alloc_strategy()+'0');
    putchar('\n');
    }

The use of the get/set_mem_alloc_strategy() functions is explained in
the previous chapter of this documentation.  When this program is run
with no command line parameter it will report the current memory
strategy value (0, 1 or 2).  You can set the memory strategy by running
the program with "0", "1" or "2" on the command line.

This program is provided mainly because it is occasionally a useful
utility.  Version 6.00 (but not 5.10) of the Microsoft C Compiler is
incompatible with memory strategy 2 (last fit).  Thus if you have been
fooling with the memory strategy, and have left it in last-fit mode, and
then try to compile a program with Version 6.00 you will get probably
get the error message:

            run-time error R6009
            - not enough space for environment

If this happens then simply run MEMSTRAT 0 to change the memory
strategy to first-fit, and the compiler will no longer complain.


(h)     Explanation of the INITDEMO.C program

The INITDEMO.C program, which shows how to allocate space for a
1-dimensional arrary of structures, is as follows:

    /*  INITDEMO.C
     *  demonstrates mem_init() and array1_alloc() functions
     *  last mod.: 14-JAN-91
     */

    #include <STDIO.H>
    #include <STDLIB.H>

    #include <L_MEM.H>

    #define M 6
    #define N ( 2+(65536/(10+M)) )

    struct date_t
       {
       int day;
       int month;
       long year;
       char foo[M];
       char calendar;
       } date;

    char msg[60];

    void main(void)
    {
    Far_str_ptr ptr1;
    struct date_t far *dptr;
    struct date_t huge *ptr2;
    int i, err_flag;

    date.day = 21;
    date.month = 12;
    date.year = 2012;
    date.calendar = 'G';

    printf("\nThis program allocates space for %d date structures.",N);
    printf("\nSize of date structure = %d bytes (M = %d).",sizeof(date),M);
    printf("\n\nFirst we use the mem_alloc() and mem_init() functions.");

    /*  allocate space for date structures  */
    ptr1 = mem_alloc(N*sizeof(struct date_t),&err_flag);
    if ( err_flag )
       {
       printf("\nFirst allocation failed: %s",mem_err_msg(err_flag,msg));
       exit(1);
       }

    mem_init(ptr1,sizeof(struct date_t),(char *)&date,&err_flag);
    if ( err_flag )
       {
       printf("\nInitialization failed: %s",mem_err_msg(err_flag,msg));
       exit(2);
       }

    /*  display last three dates in allocated space
     *  plus what is beyond last date
     */
    for ( i=N-3; i<=N; i++ )
       {
       dptr = (struct date_t far *)
           ( (Huge_str_ptr)ptr1 + (Ulong)i*sizeof(struct date_t) );
       printf("\nDate #%d = %d/%d/%ld (%c)   (dptr = %lp)",
           i,dptr->month,dptr->day,dptr->year,dptr->calendar,dptr);
       }
    printf("\nFinal \"date\" is beyond allocated space.");

    /*  now do the same thing using the array1_alloc() function  */
    printf("\n\nNow the same thing using array1_alloc()");

    ptr2 = (struct date_t huge *)array1_alloc(N,sizeof(struct date_t),
        (char *)&date, &err_flag);
    if ( err_flag )
       {
       printf("\nSecond allocation failed: %s",mem_err_msg(err_flag,msg));
       exit(3);
       }

    for ( i=N-3; i<=N; i++ )
        printf("\nDate #%d = %d/%d/%ld (%c)",
            i,ptr2[i].month,ptr2[i].day,ptr2[i].year,ptr2[i].calendar);

    printf("\nFinal \"date\" is beyond allocated space.\n");
    }

A 1-dimensional array of structures can be allocated in two ways using
functions in this library: the quick way and the slightly less quick
way.  The quick way allocates memory and (if an initialization value is
specified) initializes the memory with a single function call.  The less
quick way uses two function calls to accomplish the same task.  The
quick way allows access to the elements of the array in the usual C
manner, i.e. by use of square brackets.  Otherwise it is necessary to
achieve the same result by means of pointer arithmetic.

When INITDEMO is run, the screen output (apart from the pointer
addresses) is:

    This program allocates space for 4098 date structures.
    Size of date structure = 16 bytes (M = 6).

    First we use the mem_alloc() and mem_init() functions.
    Date #4095 = 12/21/2012 (G)   (dptr = 2C30:FFF0)
    Date #4096 = 12/21/2012 (G)   (dptr = 3C30:0000)
    Date #4097 = 12/21/2012 (G)   (dptr = 3C30:0010)
    Date #4098 = -13056/90/1663071075 ( )   (dptr = 3C30:0020)
    Final "date" is beyond allocated space.

    Now the same thing using array1_alloc()
    Date #4095 = 12/21/2012 (G)
    Date #4096 = 12/21/2012 (G)
    Date #4097 = 12/21/2012 (G)
    Date #4098 = -13824/90/141003091 ()
    Final "date" is beyond allocated space.

If memory is allocated for an array of structures, and the total amount
of memory allocated is > 64K, then the size of each element must be a
power of 2.  The reason for this is that, because of the way that the
elements of a structure are accessed, a structure should not cross a 64K
segment boundary.  For example, if a[] is a 1-dimensional array of
structures, and a has the value 5921:0000, then it is not permissible
that a structure in the array should begin at 5921:FFF0 and end at some
address past 5921:FFFF.  Thus the size of each element must divide
65,565, and so must be a power of 2.

If the value of M in INITDEMO.C is changed to 14, making the size of the
structure to be 24 instead of 16, and the program is compiled and
executed, the screen output will be:
                           
    This program allocates space for 2732 date structures.
    Size of date structure = 24 bytes (M = 14).

    First we use the mem_alloc() and mem_init() functions.
    Initialization failed: Size parameter is not a power of two.

Thus the mem_init() function, as also the array allocation functions,
will detect the case of using a non-power-of-2 size when the allocated
space exceeds 64K.


(i)     Explanation of the ARRAY3.C program

The ARRAY3.C program, in which the allocation and freeing of a
3-dimensional arrays of double precision floating point numbers is
illustrated, is as follows:

    /*  ARRAY3.C
     *  demonstrates use of memory log with 3-dimensional
     *  array allocation and free functions
     *  (also demonstrates floating point frailties)
     *  last mod.: 13-JAN-91
     */

    #include <STDIO.H>
    #include <STDLIB.H>
    #include <STRING.H>

    #include "L_MEM.H"

    #define N1 20
    #define N2 30
    #define N3 20

    int err_flag;
    char msg[100];

    typedef double huge * far * far *Array_3_double;

    void main(int argc,char **argv)
    {
    Array_3_double a1, a2, a3, a4;          /*  see typedef above  */
    double val, init_val = atof(argv[1]);   /*  a3 initialization value  */
    int i, j, k, m;
    int print_log = ( argv[2][0] == 'P' || argv[2][0] == 'p' );
                                            /*  print log if TRUE  */
    FILE *output;                           /*  log output  */

    if ( argc == 1 )
       {
       printf("Use: ARRAY3 initialization_value [P [log_file]]."
            "\n\"initialization_value\" is any floating point value."
            "\nUse \"P\" to print the memory allocation log"
            "\n(optionally to disk file \"log_file\").\n");
       exit(0);
       }

    if ( print_log )
       {
       if ( argc < 4 )
           output = stdprn;
       else if ( ( output = fopen(argv[3],"wt") ) == NULL )
          {
          printf("\nUnable to open file %s",argv[3]);
          exit(1);
          }
       }

    mem_log_on();
    /*  array_expansion_off();
     *  activatation of this statement suppresses recording in the log
     *  of memory block operations constituting array operations
     */
    strcpy(msg,"Program is ");
    strcat(msg,argv[0]);
    mem_log_msg(msg,&err_flag);

    sprintf(msg,"Initial value for a3 is %f.",init_val);
    mem_log_msg(msg,&err_flag);

    /*  set termination on detection of a critical error  */
    crit_err_termination_on();
    /* if a critical error is detected send report 1 to printer  */
    set_crit_err_termination_output(PRINTER_OUTPUT,REPORT_1_ONLY);

    /*  begin with something ridiculous  */
    array3_free(NULL,&err_flag);

    /*  now allocate three arrays  */
    printf("\nNow allocating three 3-dimensional arrays of doubles, "
           "a1, a2 and a3.");

    /*  note type casting of return value to Array_3_double,
     *  which is typedef'd above as: double huge * far * far *
     *  i.e. a far pointer to a far pointer to a huge pointer to a double
     */
    a1 = (Array_3_double)array3_alloc(N1,N2,N3,
        sizeof(double),NULL,&err_flag);
    if ( err_flag )
       {
       printf("\n%s",mem_err_msg(err_flag,msg));
       printf("\nUnable to allocate a1.\n");
       exit(1);
       }

    /*  NULL means: don't initialize the array  */
    a2 = (Array_3_double)array3_alloc(N1,N2,N3,
        sizeof(double),NULL,&err_flag);
    if ( err_flag )
       {
       printf("\n%s",mem_err_msg(err_flag,msg));
       printf("\nUnable to allocate a2.\n");
       exit(2);
       }

    /*  note initialization to given value, namely, init_val  */
    a3 = (Array_3_double)array3_alloc(N1,N2,N3,sizeof(double),
        (char *)&init_val,&err_flag);
    if ( err_flag )
       {
       printf("\n%s",mem_err_msg(err_flag,msg));
       printf("\nUnable to allocate a3.\n");
       exit(3);
       }

    /*  initialze arrays a2 and a3 subsequent to allocation  */
    printf("\nNow initializing %dx%dx%d arrays a1 and a2 ",N1,N2,N3);
    for ( i=0; i<N1; i++ )
       {
       printf("-");
       for ( j=0; j<N2; j++)
          {
          for ( k=0; k<N3; k++ )
             {
             a1[i][j][k] = (double)i + j;
             a2[i][j][k] = i - k - init_val;
             /*  check previous initialization of a3  */
             if ( a3[i][j][k] != init_val )
                {
                printf("\na3[%d][%d][%d] = %f, should be %f.",
                    i,j,k,a3[i][j][k],init_val);
                exit(1);
                }
             }
          }
       }

    printf("\nNow checking results of calculation ");
    for ( i=0; i<N1; i++ )
       {
       printf("-");
       for ( j=0; j<N2; j++)
          {
          for ( k=0; k<N3; k++ )
             {
             val = a1[i][j][k] - a2[i][j][k] - j - k;
             if ( a3[i][j][k] != val )
                {
                /*  inequality may occur due to slight inaccuracy
                 *  in representation of floating point values
                 */
                printf("\nUnexpected result:");
                printf("\na3[%d][%d][%d] = %f",i,j,k,a3[i][j][k]);
                printf("\nval = a1[%d][%d][%d] - a2[%d][%d][%d]"
                    " - %d - %d = %f",i,j,k,i,j,k,j,k,
                    a1[i][j][k]-a2[i][j][k]-j-k);
                printf("\nval - a3[i][j][k] = %f, and val %s= a3[i][j][k]",
                    val-a3[i][j][k],(val==a3[i][j][k]?"":"!"));
                /*  display the actual hexadecimal representation of
                 *  the two floating point values
                 */
                printf("\na3[%d][%d][%d]: ",i,j,k);
                for ( m=0; m<sizeof(double); m++ )
                    printf("%02X ",0xFF&*((char far *)&a3[i][j][k]+m));
                printf("\nval:         ");
                for ( m=0; m<sizeof(double); m++ )
                    printf("%02X ",0xFF&*((char far *)&val+m));
                exit(2);
                }
             }
          }
       }

    printf("\nCalculation and comparisons check out OK.\n");

    /*  check available memory (write in log)  */
    mem_max(&err_flag);

    /*  do more ridiculous things  */

    /*  attempt to allocate an impossibly large array  */
    a4 = (Array_3_double)array3_alloc(100,100,100,
        sizeof(double),NULL,&err_flag);

    /*  attempt to allocate an array of elements of size 0  */
    a4 = (Array_3_double)array3_alloc(100,100,100,0,NULL,&err_flag);

    /*  allocate a memory block  */
    mem_alloc(1000L,&err_flag);

    /*  free the first and second arrays specifically  */
    array3_free(a1,&err_flag);
    array3_free(a2,&err_flag);

    /*  attempt to free a1 and a2 a second time  */
    array3_free(a1,&err_flag);
    array3_free(a2,&err_flag);

    /*  now free allocated memory block and third array together  */
    mem_free_all(0,&err_flag);

    if ( print_log )
       {
       if ( output == stdprn )
           printf("\nNow printing reports ...");
       else
           printf("\nNow writing reports to %s.",strupr(argv[3]));
       mem_log_report_1(output,1,TRUE);
       mem_log_report_2(output,0,TRUE);
       printf("\n");
       }
    else
        printf("\nThere are %u entries in the memory allocation log."
            "\nUse \"ARRAY3 init_val P\" to print the log or "
            "\n\"ARRAY3 init_val P log_file\" to write the log to disk.\n",
            mem_log_entries());
    }

Allocation of a 3-dimensional array requires allocation of three
separate memory blocks.  The first block holds the data, the second
block is a 2-dimensional array of pointers to the rows of the first
array.  The third block is a 1-dimensional array of pointers to the rows
of pointers in the second array.  Thus each array variable, in this
example, is a pointer to a pointer to a pointer to a double.  The first
two of these pointers are far pointers, but the final pointer is a huge
pointer, since the data space (holding the double precision floating
point values) may occupy more than 64K of memory (in this case it
occupies 12000*8 bytes, or 93.75K). Thus the type of each array variable
is:

                    double huge * far * far *

In order to simplify array declarations the following typedef is used:

        typedef double huge * far * far *Array_3_double;

If 4-dimensional arrays of long ints were being used then a typedef such as:

        typedef long huge * far * far * far *Array_4_long;

might be used instead.

The program allocates three 3-dimensional arrays, a1, a2 and a3.
Because of the way that the three memory blocks are allocated, and the
two auxiliary arrays of pointers are initialized, it is possible to
refer to the elements of these arrays in the usual C style, e.g.
a[i][j][k].  The variable a is a pointer, and so can be subscripted.
a[i] is also a pointer, and so can be subscripted.  a[i][j] is also a
pointer, and so can be subscripted.  Finally, a[i][j][k] is a double,
and so we can use such statements as:

            a1[i][j][k] = (double)i + j;
            a3[i][j][k] = a1[i][j][k] - a2[i][j][k];

Notice that the array a3 is initialized at allocation, whereas the other
two arrays are initialized explicitly within a double for-loop.  a3 is
initialized by means of the statement:

    a3 = (Array_3_double)array3_alloc(N1,N2,N3,sizeof(double),
        (char *)&init_val,&err_flag);

Here init_val points to the bytes to be used for the initialization.
The number of bytes used to initialize each element of the array is
given by the fourth parameter, in this case, sizeof(double).

When run without command line paramters, the command line syntax is
displayed, as follows:

    Use: ARRAY3 initialization_value [P [log_file]].
    "initialization_value" is any floating point value.
    Use "P" to print the memory allocation log
    (optionally to disk file "log_file").

If .625 is used as the floating point initialization value for the array
a3 (as in: "ARRAY3 .625") then the screen output is:

    Now allocating three 3-dimensional arrays of doubles, a1, a2 and a3.
    Now initializing 20x30x20 arrays a1 and a2 --------------------
    Now checking results of calculation --------------------
    Calculation and comparisons check out OK.

    There are 39 entries in the memory allocation log.
    Use "ARRAY3 init_val P" to print the log or
    "ARRAY3 init_val P log_file" to write the log to disk.

If .6 is used as the floating point initialization value for the array
a3 (as in: "ARRAY3 .6") then the screen output is:

    Now allocating three 3-dimensional arrays of doubles, a1, a2 and a3.
    Now initializing 20x30x20 arrays a1 and a2 --------------------
    Now checking results of calculation -
    Unexpected result:
    a3[0][0][1] = 0.600000
    val = a1[0][0][1] - a2[0][0][1] - 0 - 1 = 0.600000
    val - a3[i][j][k] = 0.000000, and val != a3[i][j][k]
    a3[0][0][1]: 33 33 33 33 33 33 E3 3F
    val:         34 33 33 33 33 33 E3 3F

This has nothing to do with array allocation, but illustrates a
peculiarity of floating point arithmetic.  In theory, since

                     a1[i][j][k] = i + j
and                  a2[i][j][k] = i - k - init_val

it should be true that

            a1[i][j][k] - a2[i][j][k] - j - k = init_val

But the result of the arithmetic operations is an 8-byte representation
of a floating point number that differs by a single bit (as shown above)
from the 8-byte representation of init_val, and so the test for equality
fails.

A few ridiculous things are done deliberately in ARRAY3 to demonstrate
how the functions in this library handle such ridiculous things (which
in practice would, of course, only be done inadvertently).  The memory
allocation log will be something like that given below (this output to a
file called LOG is obtained with the command line "ARRAY3 .625 P LOG").
Note that each log entry for an array operation is immediately followed
by log entries giving details of the constituent memory block operations.
These are marked by ">>>". If the operation is successful there will be
three constituent memory block operations (in the case of a 3-dimensional
array, as here).

    Number of entries in the memory allocation log: 39

       1  Message: Program is E:\MEMLIB\DEMOS\ARRAY3.EXE

       2  Message: Initial value for a3 is 0.625000.

       3 "Free 3-dimensional array."
          Error: Null pointer.

       4 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.

       5 >>> "Allocate 96000 bytes."
             6000 (0x1770) paragraphs (96000 bytes) allocated at 305A:0000.

       6 >>> "Allocate 2400 bytes."
             150 (0x96) paragraphs (2400 bytes) allocated at 47CE:0000.

       7 >>> "Allocate 80 bytes."
             5 (0x5) paragraphs (80 bytes) allocated at 4868:0000.

       8 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.

       9 >>> "Allocate 96000 bytes."
             6000 (0x1770) paragraphs (96000 bytes) allocated at 4874:0000.

      10 >>> "Allocate 2400 bytes."
             150 (0x96) paragraphs (2400 bytes) allocated at 5FE8:0000.

      11 >>> "Allocate 80 bytes."
             5 (0x5) paragraphs (80 bytes) allocated at 6082:0000.

      12 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.

      13 >>> "Allocate 96000 bytes."
             6000 (0x1770) paragraphs (96000 bytes) allocated at 608E:0000.

      14 >>> "Allocate 2400 bytes."
             150 (0x96) paragraphs (2400 bytes) allocated at 7802:0000.

      15 >>> "Allocate 80 bytes."
             5 (0x5) paragraphs (80 bytes) allocated at 789C:0000.

      16 "Ascertain maximum available."  161056 bytes.

      17 "Allocate 3-dimensional array (element size = 8 bytes)."
          Data space required: 8000000 bytes.
          Error: More than 600K requested.

      18 >>> "Allocate 8000000 bytes."
             Error: More than 600K requested.

      19 "Allocate 3-dimensional array (element size = 0 bytes)."
          Error: An array parameter is zero.

      20 "Allocate 1000 bytes."
         63 (0x3F) paragraphs (1008 bytes) allocated at 78B1:0000.

      21 "Free 3-dimensional (12000-element) array."  Done.

      22 >>> "Free block at 4868:0000."  80 bytes (0x5 paragraphs) freed.

      23 >>> "Free block at 47CE:0000."  2400 bytes (0x96 paragraphs) freed.

      24 >>> "Free block at 305A:0000."  96000 bytes (0x1770 paragraphs) freed.

      25 "Free 3-dimensional (12000-element) array."  Done.

      26 >>> "Free block at 6082:0000."  80 bytes (0x5 paragraphs) freed.

      27 >>> "Free block at 5FE8:0000."  2400 bytes (0x96 paragraphs) freed.

      28 >>> "Free block at 4874:0000."  96000 bytes (0x1770 paragraphs) freed.

      29 "Free 3-dimensional (12000-element) array."
          Error: This is a memory allocation log entry.

      30 >>> "Free block at 4868:0000."
             Error: This is a memory allocation log entry.

      31 "Free 3-dimensional (12000-element) array."
          Error: Block has already been freed (or relocated).

      32 >>> "Free block at 6082:0000."
             Error: Block has already been freed (or relocated).

      33  Message: About to free all allocated blocks/arrays.

      34 "Free 3-dimensional (12000-element) array."  Done.

      35 >>> "Free block at 789C:0000."  80 bytes (0x5 paragraphs) freed.

      36 >>> "Free block at 7802:0000."  2400 bytes (0x96 paragraphs) freed.

      37 >>> "Free block at 608E:0000."  96000 bytes (0x1770 paragraphs) freed.

      38 "Free block at 78B1:0000."  1008 bytes (0x3F paragraphs) freed.

      39 "Free all blocks/arrays."  2 freed.

       Number of entries in the memory allocation log: 39

       ***  Block/array ID = 1  ***

       4 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.
      21 "Free 3-dimensional (12000-element) array."  Done.
      29 "Free 3-dimensional (12000-element) array."
          Error: This is a memory allocation log entry.

       ***  Block/array ID = 2  ***

       5 >>> "Allocate 96000 bytes."
             6000 (0x1770) paragraphs (96000 bytes) allocated at 305A:0000.
      24 >>> "Free block at 305A:0000."  96000 bytes (0x1770 paragraphs) freed.

       ***  Block/array ID = 3  ***

       6 >>> "Allocate 2400 bytes."
             150 (0x96) paragraphs (2400 bytes) allocated at 47CE:0000.
      23 >>> "Free block at 47CE:0000."  2400 bytes (0x96 paragraphs) freed.

       ***  Block/array ID = 4  ***

       7 >>> "Allocate 80 bytes."
             5 (0x5) paragraphs (80 bytes) allocated at 4868:0000.
      22 >>> "Free block at 4868:0000."  80 bytes (0x5 paragraphs) freed.

       ***  Block/array ID = 5  ***

       8 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.
      25 "Free 3-dimensional (12000-element) array."  Done.
      31 "Free 3-dimensional (12000-element) array."
          Error: Block has already been freed (or relocated).

       ***  Block/array ID = 6  ***

       9 >>> "Allocate 96000 bytes."
             6000 (0x1770) paragraphs (96000 bytes) allocated at 4874:0000.
      28 >>> "Free block at 4874:0000."  96000 bytes (0x1770 paragraphs) freed.

       ***  Block/array ID = 7  ***

      10 >>> "Allocate 2400 bytes."
             150 (0x96) paragraphs (2400 bytes) allocated at 5FE8:0000.
      27 >>> "Free block at 5FE8:0000."  2400 bytes (0x96 paragraphs) freed.

       ***  Block/array ID = 8  ***

      11 >>> "Allocate 80 bytes."
             5 (0x5) paragraphs (80 bytes) allocated at 6082:0000.
      26 >>> "Free block at 6082:0000."  80 bytes (0x5 paragraphs) freed.
      32 >>> "Free block at 6082:0000."
             Error: Block has already been freed (or relocated).

       ***  Block/array ID = 9  ***

      12 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.
      34 "Free 3-dimensional (12000-element) array."  Done.

       ***  Block/array ID = 10  ***

      13 >>> "Allocate 96000 bytes."
             6000 (0x1770) paragraphs (96000 bytes) allocated at 608E:0000.
      37 >>> "Free block at 608E:0000."  96000 bytes (0x1770 paragraphs) freed.

       ***  Block/array ID = 11  ***

      14 >>> "Allocate 2400 bytes."
             150 (0x96) paragraphs (2400 bytes) allocated at 7802:0000.
      36 >>> "Free block at 7802:0000."  2400 bytes (0x96 paragraphs) freed.

       ***  Block/array ID = 12  ***

      15 >>> "Allocate 80 bytes."
             5 (0x5) paragraphs (80 bytes) allocated at 789C:0000.
      35 >>> "Free block at 789C:0000."  80 bytes (0x5 paragraphs) freed.

       ***  Block/array ID = 15  ***

      20 "Allocate 1000 bytes."
         63 (0x3F) paragraphs (1008 bytes) allocated at 78B1:0000.
      38 "Free block at 78B1:0000."  1008 bytes (0x3F paragraphs) freed.

       Errors not specific to a memory block or array:

       3 "Free 3-dimensional array."
          Error: Null pointer.
      17 "Allocate 3-dimensional array (element size = 8 bytes)."
          Data space required: 8000000 bytes.
          Error: More than 600K requested.
      18 >>> "Allocate 8000000 bytes."
             Error: More than 600K requested.
      19 "Allocate 3-dimensional array (element size = 0 bytes)."
          Error: An array parameter is zero.

      13 memory blocks (or arrays) were allocated.
      13 blocks (or arrays) were freed.
       1 null pointer passed.

If, in the ARRAY3.C program, the statement:

                    array_expansion_off();

is activated (by moving it out of the commen delimiters) then the log
entries marked ">>>" will not be recorded in the log, i.e. the
constitutive memory block operations of each array operation will not be
recorded.  Moreover, the constituent memory blocks will not be counted
in the final summary report.  The two reports will then be as follows:

    Number of entries in the memory allocation log: 18

       1  Message: Program is E:\memlib\demos\ARRAY3.EXE

       2  Message: Initial value for a3 is 0.625000.

       3 "Free 3-dimensional array."
          Error: Null pointer.

       4 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.

       5 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.

       6 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.

       7 "Ascertain maximum available."  161488 bytes.

       8 "Allocate 3-dimensional array (element size = 8 bytes)."
          Data space required: 8000000 bytes.
          Error: More than 600K requested.

       9 "Allocate 3-dimensional array (element size = 0 bytes)."
          Error: An array parameter is zero.

      10 "Allocate 1000 bytes."
         63 (0x3F) paragraphs (1008 bytes) allocated at 7893:0000.

      11 "Free 3-dimensional (12000-element) array."  Done.

      12 "Free 3-dimensional (12000-element) array."  Done.

      13 "Free 3-dimensional (12000-element) array."
          Error: Block has already been freed (or relocated).

      14 "Free 3-dimensional (12000-element) array."
          Error: Block has already been freed (or relocated).

      15  Message: About to free all allocated blocks/arrays.

      16 "Free 3-dimensional (12000-element) array."  Done.

      17 "Free block at 7893:0000."  1008 bytes (0x3F paragraphs) freed.

      18 "Free all blocks/arrays."  2 freed.

       Number of entries in the memory allocation log: 18

       ***  Block/array ID = 1  ***

       4 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.
      11 "Free 3-dimensional (12000-element) array."  Done.
      13 "Free 3-dimensional (12000-element) array."
          Error: Block has already been freed (or relocated).

       ***  Block/array ID = 2  ***

       5 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.
      12 "Free 3-dimensional (12000-element) array."  Done.
      14 "Free 3-dimensional (12000-element) array."
          Error: Block has already been freed (or relocated).

       ***  Block/array ID = 3  ***

       6 "Allocate 3-dimensional array (element size = 8 bytes)."  Done.
         12000 elements in array; data space occupies 96000 bytes.
      16 "Free 3-dimensional (12000-element) array."  Done.

       ***  Block/array ID = 6  ***

      10 "Allocate 1000 bytes."
         63 (0x3F) paragraphs (1008 bytes) allocated at 7893:0000.
      17 "Free block at 7893:0000."  1008 bytes (0x3F paragraphs) freed.

       Errors not specific to a memory block or array:

       3 "Free 3-dimensional array."
          Error: Null pointer.
       8 "Allocate 3-dimensional array (element size = 8 bytes)."
          Data space required: 8000000 bytes.
          Error: More than 600K requested.
       9 "Allocate 3-dimensional array (element size = 0 bytes)."
          Error: An array parameter is zero.

       4 memory blocks (or arrays) were allocated.
       4 blocks (or arrays) were freed.
       1 null pointer passed.

The default condition is array expansion on.  It can be turned on and
off under programmer control.  Note that this affects the actual
recording of operations in the log, not simply their output.  If the
array_expansion_off() function is first called just prior to printing
the log then the constituent operations will still be printed, since
they have already been entered in the log.


Answer to bug identity problem in Section (b):

The statement "i--" in TINYBUG.C should have been "i -= 2".  In the
final call to mem_alloc() the index i is incremented one beyond the
index of the returned NULL pointer.  Thus in order to make i the index
of the last valid pointer it must be decremented twice, not just once.

